@@ -57,22 +57,32 @@ enum side {
 	SERVER,
 };
 
+enum visibility {
+	PRIVATE,
+	PUBLIC,
+};
+
 static int
 usage(int ret)
 {
-	fprintf(stderr, "usage: %s [OPTION] [client-header|server-header|code]"
+	fprintf(stderr, "usage: %s [OPTION] [client-header|server-header|private-code|public-code]"
 		" [input_file output_file]\n", PROGRAM_NAME);
 	fprintf(stderr, "\n");
 	fprintf(stderr, "Converts XML protocol descriptions supplied on "
 			"stdin or input file to client\n"
-			"headers, server headers, or protocol marshalling code.\n\n");
+			"headers, server headers, or protocol marshalling code.\n\n"
+			"Use \"public-code\" only if the marshalling code will be public - "
+			"aka DSO will export it while other components will be using it.\n"
+			"Using \"private-code\" is strongly recommended.\n\n");
 	fprintf(stderr, "options:\n");
 	fprintf(stderr, "    -h,  --help                  display this help and exit.\n"
 			"    -v,  --version               print the wayland library version that\n"
 			"                                 the scanner was built against.\n"
-		        "    -c,  --include-core-only     include the core version of the headers,\n"
-	                "                                 that is e.g. wayland-client-core.h instead\n"
-	                "                                 of wayland-client.h.\n");
+			"    -c,  --include-core-only     include the core version of the headers,\n"
+			"                                 that is e.g. wayland-client-core.h instead\n"
+			"                                 of wayland-client.h.\n"
+			"    -s,  --strict                exit immediately with an error if DTD\n"
+			"                                 verification fails.\n");
 	exit(ret);
 }
 
@@ -1712,9 +1722,11 @@ emit_messages(struct wl_list *message_list,
 	printf("};\n\n");
 }
 
+
 static void
-emit_code(struct protocol *protocol)
+emit_code(struct protocol *protocol, enum visibility vis)
 {
+	const char *symbol_visibility;
 	struct interface *i, *next;
 	struct wl_array types;
 	char **p, *prev;
@@ -1728,6 +1740,23 @@ emit_code(struct protocol *protocol)
 	       "#include <stdint.h>\n"
 	       "#include \"wayland-util.h\"\n\n");
 
+	/* When building a shared library symbols must be exported, otherwise
+	 * we want to have the symbols hidden. */
+	if (vis == PRIVATE) {
+		symbol_visibility = "WL_PRIVATE";
+		printf("#ifndef __has_attribute\n"
+		       "# define __has_attribute(x) 0  /* Compatibility with non-clang compilers. */\n"
+		       "#endif\n\n");
+
+		printf("#if (__has_attribute(visibility) || defined(__GNUC__) && __GNUC__ >= 4)\n"
+		       "#define WL_PRIVATE __attribute__ ((visibility(\"hidden\")))\n"
+		       "#else\n"
+		       "#define WL_PRIVATE\n"
+		       "#endif\n\n");
+	} else {
+		symbol_visibility = "WL_EXPORT";
+	}
+
 	wl_array_init(&types);
 	wl_list_for_each(i, &protocol->interface_list, link) {
 		emit_types_forward_declarations(protocol, &i->request_list, &types);
@@ -1757,10 +1786,10 @@ emit_code(struct protocol *protocol)
 		emit_messages(&i->request_list, i, "requests");
 		emit_messages(&i->event_list, i, "events");
 
-		printf("WL_EXPORT const struct wl_interface "
+		printf("%s const struct wl_interface "
 		       "%s_interface = {\n"
 		       "\t\"%s\", %d,\n",
-		       i->name, i->name, i->version);
+		       symbol_visibility, i->name, i->name, i->version);
 
 		if (!wl_list_empty(&i->request_list))
 			printf("\t%d, %s_requests,\n",
@@ -1801,11 +1830,14 @@ int main(int argc, char *argv[])
 	bool help = false;
 	bool core_headers = false;
 	bool version = false;
+	bool strict = false;
 	bool fail = false;
 	int opt;
 	enum {
 		CLIENT_HEADER,
 		SERVER_HEADER,
+		PRIVATE_CODE,
+		PUBLIC_CODE,
 		CODE,
 	} mode;
 
@@ -1813,11 +1845,12 @@ int main(int argc, char *argv[])
 		{ "help",              no_argument, NULL, 'h' },
 		{ "version",           no_argument, NULL, 'v' },
 		{ "include-core-only", no_argument, NULL, 'c' },
+		{ "strict",            no_argument, NULL, 's' },
 		{ 0,                   0,           NULL, 0 }
 	};
 
 	while (1) {
-		opt = getopt_long(argc, argv, "hvc", options, NULL);
+		opt = getopt_long(argc, argv, "hvcs", options, NULL);
 
 		if (opt == -1)
 			break;
@@ -1832,6 +1865,9 @@ int main(int argc, char *argv[])
 		case 'c':
 			core_headers = true;
 			break;
+		case 's':
+			strict = true;
+			break;
 		default:
 			fail = true;
 			break;
@@ -1853,6 +1889,10 @@ int main(int argc, char *argv[])
 		mode = CLIENT_HEADER;
 	else if (strcmp(argv[0], "server-header") == 0)
 		mode = SERVER_HEADER;
+	else if (strcmp(argv[0], "private-code") == 0)
+		mode = PRIVATE_CODE;
+	else if (strcmp(argv[0], "public-code") == 0)
+		mode = PUBLIC_CODE;
 	else if (strcmp(argv[0], "code") == 0)
 		mode = CODE;
 	else
@@ -1894,6 +1934,10 @@ int main(int argc, char *argv[])
 		"* WARNING: XML failed validation against built-in DTD *\n"
 		"*                                                     *\n"
 		"*******************************************************\n");
+		if (strict) {
+			fclose(input);
+			exit(EXIT_FAILURE);
+		}
 	}
 
 	/* create XML parser */
@@ -1936,8 +1980,17 @@ int main(int argc, char *argv[])
 		case SERVER_HEADER:
 			emit_header(&protocol, SERVER);
 			break;
+		case PRIVATE_CODE:
+			emit_code(&protocol, PRIVATE);
+			break;
 		case CODE:
-			emit_code(&protocol);
+			fprintf(stderr,
+				"Using \"code\" is deprecated - use "
+				"private-code or public-code.\n"
+				"See the help page for details.\n");
+			/* fallthrough */
+		case PUBLIC_CODE:
+			emit_code(&protocol, PUBLIC);
 			break;
 	}
 
