@@ -51,7 +51,13 @@
 
 enum wl_proxy_flag {
 	WL_PROXY_FLAG_ID_DELETED = (1 << 0),
-	WL_PROXY_FLAG_DESTROYED = (1 << 1)
+	WL_PROXY_FLAG_DESTROYED = (1 << 1),
+	WL_PROXY_FLAG_WRAPPER = (1 << 2),
+};
+
+struct wl_zombie {
+	int event_count;
+	int *fd_count;
 };
 
 struct wl_proxy {
@@ -177,7 +183,7 @@ display_protocol_error(struct wl_display *display, uint32_t code,
 		return;
 
 	/* set correct errno */
-	if (wl_interface_equal(intf, &wl_display_interface)) {
+	if (intf && wl_interface_equal(intf, &wl_display_interface)) {
 		switch (code) {
 		case WL_DISPLAY_ERROR_INVALID_OBJECT:
 		case WL_DISPLAY_ERROR_INVALID_METHOD:
@@ -222,7 +228,20 @@ wl_event_queue_init(struct wl_event_queue *queue, struct wl_display *display)
 }
 
 static void
-decrease_closure_args_refcount(struct wl_closure *closure)
+wl_proxy_unref(struct wl_proxy *proxy)
+{
+	assert(proxy->refcount > 0);
+	if (--proxy->refcount > 0)
+		return;
+
+	/* If we get here, the client must have explicitly requested
+	 * deletion. */
+	assert(proxy->flags & WL_PROXY_FLAG_DESTROYED);
+	free(proxy);
+}
+
+static void
+validate_closure_objects(struct wl_closure *closure)
 {
 	const char *signature;
 	struct argument_details arg;
@@ -237,14 +256,8 @@ decrease_closure_args_refcount(struct wl_closure *closure)
 		case 'n':
 		case 'o':
 			proxy = (struct wl_proxy *) closure->args[i].o;
-			if (proxy) {
-				if (proxy->flags & WL_PROXY_FLAG_DESTROYED)
-					closure->args[i].o = NULL;
-
-				proxy->refcount--;
-				if (!proxy->refcount)
-					free(proxy);
-			}
+			if (proxy && proxy->flags & WL_PROXY_FLAG_DESTROYED)
+				closure->args[i].o = NULL;
 			break;
 		default:
 			break;
@@ -252,31 +265,47 @@ decrease_closure_args_refcount(struct wl_closure *closure)
 	}
 }
 
-void
-proxy_destroy(struct wl_proxy *proxy);
+/* Destroys a closure which was demarshaled for dispatch; unrefs all the
+ * proxies in its arguments, as well as its own proxy, and destroys the
+ * closure itself. */
+static void
+destroy_queued_closure(struct wl_closure *closure)
+{
+	const char *signature;
+	struct argument_details arg;
+	struct wl_proxy *proxy;
+	int i, count;
+
+	signature = closure->message->signature;
+	count = arg_count_for_signature(signature);
+	for (i = 0; i < count; i++) {
+		signature = get_next_argument(signature, &arg);
+		switch (arg.type) {
+		case 'n':
+		case 'o':
+			proxy = (struct wl_proxy *) closure->args[i].o;
+			if (proxy)
+				wl_proxy_unref(proxy);
+			break;
+		default:
+			break;
+		}
+	}
+
+	wl_proxy_unref(closure->proxy);
+	wl_closure_destroy(closure);
+}
 
 static void
 wl_event_queue_release(struct wl_event_queue *queue)
 {
 	struct wl_closure *closure;
-	struct wl_proxy *proxy;
-	bool proxy_destroyed;
 
 	while (!wl_list_empty(&queue->event_list)) {
 		closure = container_of(queue->event_list.next,
 				       struct wl_closure, link);
 		wl_list_remove(&closure->link);
-
-		decrease_closure_args_refcount(closure);
-
-		proxy = closure->proxy;
-		proxy_destroyed = !!(proxy->flags & WL_PROXY_FLAG_DESTROYED);
-
-		proxy->refcount--;
-		if (proxy_destroyed && !proxy->refcount)
-			free(proxy);
-
-		wl_closure_destroy(closure);
+		destroy_queued_closure(closure);
 	}
 }
 
@@ -326,6 +355,66 @@ wl_display_create_queue(struct wl_display *display)
 	return queue;
 }
 
+static int
+message_count_fds(const char *signature)
+{
+	unsigned int count, i, fds = 0;
+	struct argument_details arg;
+
+	count = arg_count_for_signature(signature);
+	for (i = 0; i < count; i++) {
+		signature = get_next_argument(signature, &arg);
+		if (arg.type == 'h')
+			fds++;
+	}
+
+	return fds;
+}
+
+static struct wl_zombie *
+prepare_zombie(struct wl_proxy *proxy)
+{
+	const struct wl_interface *interface = proxy->object.interface;
+	const struct wl_message *message;
+	int i, count;
+	struct wl_zombie *zombie = NULL;
+
+	/* If we hit an event with an FD, ensure we have a zombie object and
+	 * fill the fd_count slot for that event with the number of FDs for
+	 * that event. Interfaces with no events containing FDs will not have
+	 * zombie objects created. */
+	for (i = 0; i < interface->event_count; i++) {
+		message = &interface->events[i];
+		count = message_count_fds(message->signature);
+
+		if (!count)
+			continue;
+
+		if (!zombie) {
+			zombie = zalloc(sizeof(*zombie) +
+				        (interface->event_count * sizeof(int)));
+			if (!zombie)
+				return NULL;
+
+			zombie->event_count = interface->event_count;
+			zombie->fd_count = (int *) &zombie[1];
+		}
+
+		zombie->fd_count[i] = count;
+	}
+
+	return zombie;
+}
+
+static enum wl_iterator_result
+free_zombies(void *element, void *data, uint32_t flags)
+{
+	if (flags & WL_MAP_ENTRY_ZOMBIE)
+		free(element);
+
+	return WL_ITERATOR_CONTINUE;
+}
+
 static struct wl_proxy *
 proxy_create(struct wl_proxy *factory, const struct wl_interface *interface,
 	     uint32_t version)
@@ -404,30 +493,36 @@ wl_proxy_create_for_id(struct wl_proxy *factory,
 	return proxy;
 }
 
-void
+static void
 proxy_destroy(struct wl_proxy *proxy)
 {
-	if (proxy->flags & WL_PROXY_FLAG_ID_DELETED)
+	if (proxy->flags & WL_PROXY_FLAG_ID_DELETED) {
 		wl_map_remove(&proxy->display->objects, proxy->object.id);
-	else if (proxy->object.id < WL_SERVER_ID_START)
-		wl_map_insert_at(&proxy->display->objects, 0,
-				 proxy->object.id, WL_ZOMBIE_OBJECT);
-	else
+	} else if (proxy->object.id < WL_SERVER_ID_START) {
+		struct wl_zombie *zombie = prepare_zombie(proxy);
+
+		/* The map now contains the zombie entry, until the delete_id
+		 * event arrives. */
+		wl_map_insert_at(&proxy->display->objects,
+				 WL_MAP_ENTRY_ZOMBIE,
+				 proxy->object.id,
+				 zombie);
+	} else {
 		wl_map_insert_at(&proxy->display->objects, 0,
 				 proxy->object.id, NULL);
-
+	}
 
 	proxy->flags |= WL_PROXY_FLAG_DESTROYED;
 
-	proxy->refcount--;
-	if (!proxy->refcount)
-		free(proxy);
+	wl_proxy_unref(proxy);
 }
 
 /** Destroy a proxy object
  *
  * \param proxy The proxy to be destroyed
  *
+ * \c proxy must not be a proxy wrapper.
+ *
  * \memberof wl_proxy
  */
 WL_EXPORT void
@@ -435,6 +530,9 @@ wl_proxy_destroy(struct wl_proxy *proxy)
 {
 	struct wl_display *display = proxy->display;
 
+	if (proxy->flags & WL_PROXY_FLAG_WRAPPER)
+		wl_abort("Tried to destroy wrapper with wl_proxy_destroy()\n");
+
 	pthread_mutex_lock(&display->mutex);
 	proxy_destroy(proxy);
 	pthread_mutex_unlock(&display->mutex);
@@ -455,12 +553,17 @@ wl_proxy_destroy(struct wl_proxy *proxy)
  * \c n, \c implementation[n] should point to the handler of \c n for
  * the given object.
  *
+ * \c proxy must not be a proxy wrapper.
+ *
  * \memberof wl_proxy
  */
 WL_EXPORT int
 wl_proxy_add_listener(struct wl_proxy *proxy,
 		      void (**implementation)(void), void *data)
 {
+	if (proxy->flags & WL_PROXY_FLAG_WRAPPER)
+		wl_abort("Proxy %p is a wrapper\n", proxy);
+
 	if (proxy->object.implementation || proxy->dispatcher) {
 		wl_log("proxy %p already has listener\n", proxy);
 		return -1;
@@ -507,6 +610,8 @@ wl_proxy_get_listener(struct wl_proxy *proxy)
  * The exact details of dispatcher_data depend on the dispatcher used.  This
  * function is intended to be used by language bindings, not user code.
  *
+ * \c proxy must not be a proxy wrapper.
+ *
  * \memberof wl_proxy
  */
 WL_EXPORT int
@@ -514,6 +619,9 @@ wl_proxy_add_dispatcher(struct wl_proxy *proxy,
 			wl_dispatcher_func_t dispatcher,
 			const void *implementation, void *data)
 {
+	if (proxy->flags & WL_PROXY_FLAG_WRAPPER)
+		wl_abort("Proxy %p is a wrapper\n", proxy);
+
 	if (proxy->object.implementation || proxy->dispatcher) {
 		wl_log("proxy %p already has listener\n", proxy);
 		return -1;
@@ -568,7 +676,7 @@ create_outgoing_proxy(struct wl_proxy *proxy, const struct wl_message *message,
  *
  * For new-id arguments, this function will allocate a new wl_proxy
  * and send the ID to the server.  The new wl_proxy will be returned
- * on success or NULL on errror with errno set accordingly.  The newly
+ * on success or NULL on error with errno set accordingly.  The newly
  * created proxy will inherit their version from their parent.
  *
  * \note This is intended to be used by language bindings and not in
@@ -603,7 +711,7 @@ wl_proxy_marshal_array_constructor(struct wl_proxy *proxy,
  *
  * For new-id arguments, this function will allocate a new wl_proxy
  * and send the ID to the server.  The new wl_proxy will be returned
- * on success or NULL on errror with errno set accordingly.  The newly
+ * on success or NULL on error with errno set accordingly.  The newly
  * created proxy will have the version specified.
  *
  * \note This is intended to be used by language bindings and not in
@@ -698,7 +806,7 @@ wl_proxy_marshal(struct wl_proxy *proxy, uint32_t opcode, ...)
  *
  * For new-id arguments, this function will allocate a new wl_proxy
  * and send the ID to the server.  The new wl_proxy will be returned
- * on success or NULL on errror with errno set accordingly.  The newly
+ * on success or NULL on error with errno set accordingly.  The newly
  * created proxy will inherit their version from their parent.
  *
  * \note This should not normally be used by non-generated code.
@@ -736,7 +844,7 @@ wl_proxy_marshal_constructor(struct wl_proxy *proxy, uint32_t opcode,
  *
  * For new-id arguments, this function will allocate a new wl_proxy
  * and send the ID to the server.  The new wl_proxy will be returned
- * on success or NULL on errror with errno set accordingly.  The newly
+ * on success or NULL on error with errno set accordingly.  The newly
  * created proxy will have the version specified.
  *
  * \note This should not normally be used by non-generated code.
@@ -790,12 +898,26 @@ display_handle_error(void *data,
 		     uint32_t code, const char *message)
 {
 	struct wl_proxy *proxy = object;
+	uint32_t object_id;
+	const struct wl_interface *interface;
 
-	wl_log("%s@%u: error %d: %s\n",
-	       proxy->object.interface->name, proxy->object.id, code, message);
+	if (proxy) {
+		wl_log("%s@%u: error %d: %s\n",
+		       proxy->object.interface->name,
+		       proxy->object.id,
+		       code, message);
 
-	display_protocol_error(display, code, proxy->object.id,
-			       proxy->object.interface);
+		object_id = proxy->object.id;
+		interface = proxy->object.interface;
+	} else {
+		wl_log("[destroyed object]: error %d: %s\n",
+		       code, message);
+
+		object_id = 0;
+		interface = NULL;
+	}
+
+	display_protocol_error(display, code, object_id, interface);
 }
 
 static void
@@ -807,13 +929,16 @@ display_handle_delete_id(void *data, struct wl_display *display, uint32_t id)
 
 	proxy = wl_map_lookup(&display->objects, id);
 
-	if (!proxy)
-		wl_log("error: received delete_id for unknown id (%u)\n", id);
-
-	if (proxy && proxy != WL_ZOMBIE_OBJECT)
-		proxy->flags |= WL_PROXY_FLAG_ID_DELETED;
-	else
+	if (wl_object_is_zombie(&display->objects, id)) {
+		/* For zombie objects, the 'proxy' is actually the zombie
+		 * event-information structure, which we can free. */
+		free(proxy);
 		wl_map_remove(&display->objects, id);
+	} else if (proxy) {
+		proxy->flags |= WL_PROXY_FLAG_ID_DELETED;
+	} else {
+		wl_log("error: received delete_id for unknown id (%u)\n", id);
+	}
 
 	pthread_mutex_unlock(&display->mutex);
 }
@@ -830,9 +955,17 @@ connect_to_socket(const char *name)
 	socklen_t size;
 	const char *runtime_dir;
 	int name_size, fd;
+	bool path_is_absolute;
+
+	if (name == NULL)
+		name = getenv("WAYLAND_DISPLAY");
+	if (name == NULL)
+		name = "wayland-0";
+
+	path_is_absolute = name[0] == '/';
 
 	runtime_dir = getenv("XDG_RUNTIME_DIR");
-	if (!runtime_dir) {
+	if (!runtime_dir && !path_is_absolute) {
 		wl_log("error: XDG_RUNTIME_DIR not set in the environment.\n");
 		/* to prevent programs reporting
 		 * "failed to create display: Success" */
@@ -840,25 +973,32 @@ connect_to_socket(const char *name)
 		return -1;
 	}
 
-	if (name == NULL)
-		name = getenv("WAYLAND_DISPLAY");
-	if (name == NULL)
-		name = "wayland-0";
-
 	fd = wl_os_socket_cloexec(PF_LOCAL, SOCK_STREAM, 0);
 	if (fd < 0)
 		return -1;
 
 	memset(&addr, 0, sizeof addr);
 	addr.sun_family = AF_LOCAL;
-	name_size =
-		snprintf(addr.sun_path, sizeof addr.sun_path,
-			 "%s/%s", runtime_dir, name) + 1;
+	if (!path_is_absolute) {
+		name_size =
+			snprintf(addr.sun_path, sizeof addr.sun_path,
+			         "%s/%s", runtime_dir, name) + 1;
+	} else {
+		/* absolute path */
+		name_size =
+			snprintf(addr.sun_path, sizeof addr.sun_path,
+			         "%s", name) + 1;
+	}
 
 	assert(name_size > 0);
 	if (name_size > (int)sizeof addr.sun_path) {
-		wl_log("error: socket path \"%s/%s\" plus null terminator"
-		       " exceeds 108 bytes\n", runtime_dir, name);
+		if (!path_is_absolute) {
+			wl_log("error: socket path \"%s/%s\" plus null terminator"
+			       " exceeds %i bytes\n", runtime_dir, name, (int) sizeof(addr.sun_path));
+		} else {
+			wl_log("error: socket path \"%s\" plus null terminator"
+			       " exceeds %i bytes\n", name, (int) sizeof(addr.sun_path));
+		}
 		close(fd);
 		/* to prevent programs reporting
 		 * "failed to add socket: Success" */
@@ -967,6 +1107,16 @@ wl_display_connect_to_fd(int fd)
  * its value will be replaced with the WAYLAND_DISPLAY environment
  * variable if it is set, otherwise display "wayland-0" will be used.
  *
+ * If \c name is an absolute path, then that path is used as-is for
+ * the location of the socket at which the Wayland server is listening;
+ * no qualification inside XDG_RUNTIME_DIR is attempted.
+ *
+ * If \c name is \c NULL and the WAYLAND_DISPLAY environment variable
+ * is set to an absolute pathname, then that pathname is used as-is
+ * for the socket in the same manner as if \c name held an absolute
+ * path. Support for absolute paths in \c name and WAYLAND_DISPLAY
+ * is present since Wayland version 1.15.
+ *
  * \memberof wl_display
  */
 WL_EXPORT struct wl_display *
@@ -979,7 +1129,7 @@ wl_display_connect(const char *name)
 	if (connection) {
 		int prev_errno = errno;
 		errno = 0;
-		fd = strtol(connection, &end, 0);
+		fd = strtol(connection, &end, 10);
 		if (errno != 0 || connection == end || *end != '\0')
 			return NULL;
 		errno = prev_errno;
@@ -1010,6 +1160,7 @@ WL_EXPORT void
 wl_display_disconnect(struct wl_display *display)
 {
 	wl_connection_destroy(display->connection);
+	wl_map_for_each(&display->objects, free_zombies, NULL);
 	wl_map_release(&display->objects);
 	wl_event_queue_release(&display->default_queue);
 	wl_event_queue_release(&display->display_queue);
@@ -1072,14 +1223,23 @@ static const struct wl_callback_listener sync_listener = {
 WL_EXPORT int
 wl_display_roundtrip_queue(struct wl_display *display, struct wl_event_queue *queue)
 {
+	struct wl_display *display_wrapper;
 	struct wl_callback *callback;
 	int done, ret = 0;
 
 	done = 0;
-	callback = wl_display_sync(display);
+
+	display_wrapper = wl_proxy_create_wrapper(display);
+	if (!display_wrapper)
+		return -1;
+
+	wl_proxy_set_queue((struct wl_proxy *) display_wrapper, queue);
+	callback = wl_display_sync(display_wrapper);
+	wl_proxy_wrapper_destroy(display_wrapper);
+
 	if (callback == NULL)
 		return -1;
-	wl_proxy_set_queue((struct wl_proxy *) callback, queue);
+
 	wl_callback_add_listener(callback, &sync_listener, &done);
 	while (!done && ret >= 0)
 		ret = wl_display_dispatch_queue(display, queue);
@@ -1172,6 +1332,8 @@ increase_closure_args_refcount(struct wl_closure *closure)
 			break;
 		}
 	}
+
+	closure->proxy->refcount++;
 }
 
 static int
@@ -1191,11 +1353,16 @@ queue_event(struct wl_display *display, int len)
 	if (len < size)
 		return 0;
 
+	/* If our proxy is gone or a zombie, just eat the event (and any FDs,
+	 * if applicable). */
 	proxy = wl_map_lookup(&display->objects, id);
-	if (proxy == WL_ZOMBIE_OBJECT) {
-		wl_connection_consume(display->connection, size);
-		return size;
-	} else if (proxy == NULL) {
+	if (!proxy || wl_object_is_zombie(&display->objects, id)) {
+		struct wl_zombie *zombie = wl_map_lookup(&display->objects, id);
+
+		if (zombie && zombie->fd_count[opcode])
+			wl_connection_close_fds_in(display->connection,
+						   zombie->fd_count[opcode]);
+
 		wl_connection_consume(display->connection, size);
 		return size;
 	}
@@ -1216,9 +1383,8 @@ queue_event(struct wl_display *display, int len)
 		return -1;
 	}
 
-	increase_closure_args_refcount(closure);
-	proxy->refcount++;
 	closure->proxy = proxy;
+	increase_closure_args_refcount(closure);
 
 	if (proxy == &display->proxy)
 		queue = &display->display_queue;
@@ -1245,17 +1411,11 @@ dispatch_event(struct wl_display *display, struct wl_event_queue *queue)
 
 	/* Verify that the receiving object is still valid by checking if has
 	 * been destroyed by the application. */
-
-	decrease_closure_args_refcount(closure);
+	validate_closure_objects(closure);
 	proxy = closure->proxy;
 	proxy_destroyed = !!(proxy->flags & WL_PROXY_FLAG_DESTROYED);
-
-	proxy->refcount--;
 	if (proxy_destroyed) {
-		if (!proxy->refcount)
-			free(proxy);
-
-		wl_closure_destroy(closure);
+		destroy_queued_closure(closure);
 		return;
 	}
 
@@ -1275,9 +1435,9 @@ dispatch_event(struct wl_display *display, struct wl_event_queue *queue)
 				  &proxy->object, opcode, proxy->user_data);
 	}
 
-	wl_closure_destroy(closure);
-
 	pthread_mutex_lock(&display->mutex);
+
+	destroy_queued_closure(closure);
 }
 
 static int
@@ -1573,7 +1733,7 @@ wl_display_poll(struct wl_display *display, short int events)
  * the appropriate event queues. Finally, events on given event queue are
  * dispatched. On failure -1 is returned and errno set appropriately.
  *
- * In a multi threaded enviroment, do not manually wait using poll() (or
+ * In a multi threaded environment, do not manually wait using poll() (or
  * equivalent) before calling this function, as doing so might cause a dead
  * lock. If external reliance on poll() (or equivalent) is required, see
  * wl_display_prepare_read_queue() of how to do so.
@@ -1678,7 +1838,7 @@ wl_display_dispatch_queue_pending(struct wl_display *display,
  * the appropriate event queues. Finally, events on the default event queue
  * are dispatched. On failure -1 is returned and errno set appropriately.
  *
- * In a multi threaded enviroment, do not manually wait using poll() (or
+ * In a multi threaded environment, do not manually wait using poll() (or
  * equivalent) before calling this function, as doing so might cause a dead
  * lock. If external reliance on poll() (or equivalent) is required, see
  * wl_display_prepare_read_queue() of how to do so.
@@ -1756,10 +1916,12 @@ wl_display_get_error(struct wl_display *display)
 /** Retrieves the information about a protocol error:
  *
  * \param display    The Wayland display
- * \param interface  if not NULL, stores the interface where the error occurred
+ * \param interface  if not NULL, stores the interface where the error occurred,
+ *                   or NULL, if unknown.
  * \param id         if not NULL, stores the object id that generated
- *                   the error. There's no guarantee the object is
- *                   still valid; the client must know if it deleted the object.
+ *                   the error, or 0, if the object id is unknown. There's no
+ *                   guarantee the object is still valid; the client must know
+ *                   if it deleted the object.
  * \return           The error code as defined in the interface specification.
  *
  * \code
@@ -1939,6 +2101,98 @@ wl_proxy_set_queue(struct wl_proxy *proxy, struct wl_event_queue *queue)
 		proxy->queue = &proxy->display->default_queue;
 }
 
+/** Create a proxy wrapper for making queue assignments thread-safe
+ *
+ * \param proxy The proxy object to be wrapped
+ * \return A proxy wrapper for the given proxy or NULL on failure
+ *
+ * A proxy wrapper is type of 'struct wl_proxy' instance that can be used when
+ * sending requests instead of using the original proxy. A proxy wrapper does
+ * not have an implementation or dispatcher, and events received on the
+ * object is still emitted on the original proxy. Trying to set an
+ * implementation or dispatcher will have no effect but result in a warning
+ * being logged.
+ *
+ * Setting the proxy queue of the proxy wrapper will make new objects created
+ * using the proxy wrapper use the set proxy queue.
+ * Even though there is no implementation nor dispatcher, the proxy queue can
+ * be changed. This will affect the default queue of new objects created by
+ * requests sent via the proxy wrapper.
+ *
+ * A proxy wrapper can only be destroyed using wl_proxy_wrapper_destroy().
+ *
+ * A proxy wrapper must be destroyed before the proxy it was created from.
+ *
+ * If a user reads and dispatches events on more than one thread, it is
+ * necessary to use a proxy wrapper when sending requests on objects when the
+ * intention is that a newly created proxy is to use a proxy queue different
+ * from the proxy the request was sent on, as creating the new proxy and then
+ * setting the queue is not thread safe.
+ *
+ * For example, a module that runs using its own proxy queue that needs to
+ * do display roundtrip must wrap the wl_display proxy object before sending
+ * the wl_display.sync request. For example:
+ *
+ * \code
+ *
+ *   struct wl_event_queue *queue = ...;
+ *   struct wl_display *wrapped_display;
+ *   struct wl_callback *callback;
+ *
+ *   wrapped_display = wl_proxy_create_wrapper(display);
+ *   wl_proxy_set_queue((struct wl_proxy *) wrapped_display, queue);
+ *   callback = wl_display_sync(wrapped_display);
+ *   wl_proxy_wrapper_destroy(wrapped_display);
+ *   wl_callback_add_listener(callback, ...);
+ *
+ * \endcode
+ *
+ * \memberof wl_proxy
+ */
+WL_EXPORT void *
+wl_proxy_create_wrapper(void *proxy)
+{
+	struct wl_proxy *wrapped_proxy = proxy;
+	struct wl_proxy *wrapper;
+
+	wrapper = zalloc(sizeof *wrapper);
+	if (!wrapper)
+		return NULL;
+
+	pthread_mutex_lock(&wrapped_proxy->display->mutex);
+
+	wrapper->object.interface = wrapped_proxy->object.interface;
+	wrapper->object.id = wrapped_proxy->object.id;
+	wrapper->version = wrapped_proxy->version;
+	wrapper->display = wrapped_proxy->display;
+	wrapper->queue = wrapped_proxy->queue;
+	wrapper->flags = WL_PROXY_FLAG_WRAPPER;
+	wrapper->refcount = 1;
+
+	pthread_mutex_unlock(&wrapped_proxy->display->mutex);
+
+	return wrapper;
+}
+
+/** Destroy a proxy wrapper
+ * \param proxy_wrapper The proxy wrapper to be destroyed
+ *
+ * \memberof wl_proxy
+ */
+WL_EXPORT void
+wl_proxy_wrapper_destroy(void *proxy_wrapper)
+{
+	struct wl_proxy *wrapper = proxy_wrapper;
+
+	if (!(wrapper->flags & WL_PROXY_FLAG_WRAPPER))
+		wl_abort("Tried to destroy non-wrapper proxy with "
+			 "wl_proxy_wrapper_destroy\n");
+
+	assert(wrapper->refcount == 1);
+
+	free(wrapper);
+}
+
 WL_EXPORT void
 wl_log_set_handler_client(wl_log_func_t handler)
 {
